/*
b * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.aurorasoftworks.signal.runtime.core.context;


import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import com.aurorasoftworks.signal.runtime.core.context.proxy.IProxyFactory;
import com.aurorasoftworks.signal.runtime.core.context.proxy.ProxyFactory;



/**
 * A default implementation of the {@link IContext} interface.
 * This class is thread-safe.
 * 
 * @author Marek
 */
public class Context implements IContext
{
	private static IProxyFactory defaultProxyFactory
			= ProxyFactory.getInstance();
	
	public static synchronized IProxyFactory getDefaultProxyFactory()
	{
		return defaultProxyFactory;
	}

	public static synchronized void setDefaultProxyFactory(
			IProxyFactory defaultProxyFactory)
	{
		Context.defaultProxyFactory = defaultProxyFactory;
	}
	
	/**
	 * A map of beans registered in this context. Bean names ({@link String})
	 * are used as map keys and bean references ({@link Object})
	 * are used as values.
	 */
	private Hashtable beans;
	
	private IProxyFactory proxyFactory;
	
	/**
	 * Creates a new {@link Context} instance.
	 */
	public Context()
	{
		beans = new Hashtable();
	}

	/**
	 * {@inheritDoc}
	 */
	public synchronized boolean containsBean(String name)
	{
		return beans.containsKey(name);
	}

	/**
	 * {@inheritDoc}
	 */
	public synchronized Object getBean(String name) throws ContextException
	{
		IBeanReference ref = getBeanReference(name);
		return (ref == null ? null : ref.getBean());
	}

	/**
	 * {@inheritDoc}
	 */
	protected IBeanReference getBeanReference(String name)
	{
		return (IBeanReference) beans.get(name);
	}

	/**
	 * {@inheritDoc}
	 */
	public synchronized Vector getBeanNames(Class beanClass)
			throws ContextException
	{
		Vector result = new Vector();
		
		for (Enumeration en = beans.keys(); en.hasMoreElements();)
		{
			String name = (String) en.nextElement();
			if (beanClass.isAssignableFrom(getBean(name).getClass()))
			{
				result.addElement(name);
			}
		}
		
		return result;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public synchronized Vector getBeans(Class beanClass)
			throws ContextException
	{
		Vector result = new Vector();
		
		for (Enumeration en = beans.elements(); en.hasMoreElements();)
		{
			IBeanReference ref = (IBeanReference) en.nextElement();

			if (beanClass.isAssignableFrom(ref.getBeanClass()))
			{
				result.addElement(ref.getBean());
			}
		}
		
		return result;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public Object getBean(Class beanClass) throws ContextException
	{
		/* The call to getBeans is synchronized, so there is
		 * no need to synchronize this method */
		Vector results = getBeans(beanClass);
		
		if (results.size() == 0)
		{
			throw new ContextException("Beans of type "
					+ beanClass + " not found");
		}
		else if(results.size() > 1)
		{
			throw new ContextException("Multiple beans of type "
					+ beanClass + " found");
		}
		else
		{
			return results.elementAt(0);
		}
	}
	
	public Object processLazyBean(Object processedResult)
			throws ContextException
	{
		Object result;
		
		if (processedResult instanceof IBeanReference)
		{
			result = ((IBeanReference) processedResult).getBean();
		}
		else
		{
			result = processedResult;
		}
		
		return result;
	}
	
	/**
	 * Handles a bean or a bean reference returned by a
	 * {@link IBeanProcessor}.
	 * 
	 * @param ref				a bean reference passed to
	 * 							a {@link IBeanProcessor}
	 * @param processedResult   a result returned by a {@link IBeanProcessor}
	 * 							(a bean or a {@link IBeanReference} object)
	 * @throws ContextException if an error is encountered while registering
	 *                          a new bean or a new bean reference
	 */
	protected void processBean(IBeanReference ref, Object processedResult)
			throws ContextException
	{
		/* If the returned object is different from the passed parameter */
		if (processedResult != ref)
		{
			/* A new reference was returned */
			if (processedResult instanceof IBeanReference)
			{
				registerBeanReference((IBeanReference) processedResult);
			}
			/* A regular bean was returned */
			else
			{
				doRegisterBean(ref.getName(), processedResult);
			}
		}
	}
	
	/**
	 * {@inheritDoc}
	 */
	public synchronized void registerBean(String name, Object bean)
			throws ContextException
	{
		doRegisterBean(name, bean);
	}

	protected void doRegisterBean(String name, Object bean)
			throws ContextException
	{
		registerBeanReference(new EagerInitBeanReference(name, bean));
	}
	
	/**
	 * Registers a bean reference. Refer to {@link IBeanReference}
	 * for more information.
	 * 
	 * @param ref a bean reference to be registered
	 * @throws ContextException 
	 */
	protected void registerBeanReference(IBeanReference ref)
			throws ContextException
	{
		beans.put(ref.getName(), ref);

		if (IProxyFactory.class.isAssignableFrom(ref.getBeanClass()))
		{
			proxyFactory = (IProxyFactory) ref.getBean();
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public synchronized IProxyFactory getProxyFactory()
	{
		if (proxyFactory == null)
		{
			proxyFactory = getDefaultProxyFactory();
		}

		return proxyFactory;
	}
}
