/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.aurorasoftworks.signal.runtime.core.context;


/**
 * An implementation of the {@link IContextLoader} interface that instantiates
 * {@link IContext} objects with reflection.
 * 
 * 
 * @author Marek
 */
public class ContextClassLoader implements IContextLoader
{
	/**
	 * A singleton instance of this class.
	 */
	private static final IContextLoader INSTANCE = new ContextClassLoader();
	
	/**
	 * Returns the singleton instance of this class.
	 * 
	 * @return singleton instance
	 */
	public static IContextLoader getInstance()
	{
		return INSTANCE;
	}
	
	/**
	 * Creates an instance of {@link ContextClassLoader}. This constructor
	 * is declared as protected rather than private so that this 
	 * class can be extended if needed.
	 */
	protected ContextClassLoader()
	{
	}
	
	/**
	 * Loads a context by instantiating the specified class.
	 * 
	 * @param contextClass  name of a context class to be instantiated
	 * @return a loaded context
	 * @throws ContextException if an error occurs while loading the context
	 */
	public IContext loadContext(String contextClass)
			throws ContextException
	{
		return createContextInstance(loadContextClass(contextClass));
	}

	/**
	 * Creates a context instance.
	 * 
	 * @param contextClassName
	 * @return a new instance of the provided context class
	 * @throws ContextException
	 */
	protected IContext createContextInstance(Class contextClassName)
			throws ContextException
	{
		Object obj;
		try
		{
			obj = contextClassName.newInstance();
		}
		catch (Exception e)
		{
			throw new ContextException(
					"Failed to create a context instance", e);
		}
		
		try
		{
			return (IContext) obj; 
		}
		catch (ClassCastException e)
		{
			throw new ContextException(obj.getClass().getName()
					+ " does not implement " + IContext.class.getName(), e);
		}
	}

	/**
	 * Loads the specified context class.
	 * 
	 * @param contextClassName name of a class to be loaded
	 * @return loaded class
	 * @throws ContextException if the class cannot be loaded
	 */
	protected Class loadContextClass(String contextClassName)
			throws ContextException
	{
		try
		{
			return Class.forName(contextClassName);
		}
		catch (ClassNotFoundException e)
		{
			throw new ContextException(
					"Failed to load a context class", e);
		}
	}
}
