/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.aurorasoftworks.signal.runtime.core.context;


/**
 * A reference to a lazily-initialized bean. This class acts
 * like a bean factory that creates a bean when {@link #getBean()}
 * is invoked for the first time.
 * 
 * Note: methods of this class are not thread-safe because it is assumed
 * that the calling context is responsible for synchronization.
 * 
 * @author Marek
 */
public abstract class LazyInitBeanReference extends AbstractBeanReference
{
	private Class beanClass;
	
	/**
	 * Creates a new instance of LazyInitBeanReference.
	 * The wrapped bean is later created by the {@link #createBean()}
	 * method. 
	 * 
	 * @param name      name of the wrapped bean
	 * @param beanClass class of the wrapped bean
	 */
	public LazyInitBeanReference(String name, Class beanClass)
	{
		super(name);
		this.beanClass = beanClass;
	}

	/**
	 * {@inheritDoc}
	 */
	public Object getBean() throws ContextException
	{
		Object bean = super.getBean();
		
		if (bean == null)
		{
			try
			{
				bean = createBean();
			}
			catch (Exception e)
			{
				throw new ContextException(
						"Failed to create bean: " + getName(), e);
			}
			setBean(bean);
		}
		
		return bean;
	}
	
	/**
	 * Creates a wrapped bean on demand. This method is only invoked
	 * once when {@link #getBean()} is executed for the first time.
	 * 
	 * @return wrapped bean
	 * @throws Exception if an error occurs while creating the bean
	 */
	protected abstract Object createBean() throws Exception;

	public Class getBeanClass()
	{
		return beanClass;
	}
}
