/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.aurorasoftworks.signal.runtime.core.context;


import static org.testng.AssertJUnit.assertEquals;
import static org.testng.AssertJUnit.assertFalse;
import static org.testng.AssertJUnit.assertNotNull;
import static org.testng.AssertJUnit.assertNull;
import static org.testng.AssertJUnit.assertSame;
import static org.testng.AssertJUnit.assertTrue;

import java.util.Hashtable;

import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import com.aurorasoftworks.signal.runtime.core.context.proxy.IProxyFactory;
import com.aurorasoftworks.signal.runtime.core.context.proxy.ProxyFactory;

import edu.emory.mathcs.backport.java.util.Arrays;


@Test
public class ContextTest
{
	private IContext ctx;
	
	private static IContext sharedContext = new Context();
	
	@BeforeMethod
	protected void setUp()
	{
		ctx = new Context();
	}
	
	public void testRegisterBean() throws ContextException
	{
		final String name = "my-bean";
		final Object bean = new Object();
		
		assertFalse(ctx.containsBean(name));
		assertNull(ctx.getBean(name));
		
		ctx.registerBean(name, bean);
		assertTrue(ctx.containsBean(name));
		assertSame(bean, ctx.getBean(name));
	}
	
	@Test(invocationCount = 100, threadPoolSize = 4)
	public void testConcurrency() throws ContextException
	{
		final String name = "my-bean";
		final Object bean = new Object();
		
		sharedContext.registerBean(name, bean);
		assertNotNull(sharedContext.getBean(name));
	}
	
	public void testGetBeanNames() throws ContextException
	{
		final String name1 = "bean1";
		final Object obj1 = new String();
		final String name2 = "bean2";
		final Object obj2 = new Hashtable();
		
		ctx.registerBean(name1, obj1);
		ctx.registerBean(name2, obj2);
		
		assertEquals(Arrays.asList(new String[]{name2}),
				ctx.getBeanNames(Hashtable.class));
	}
	
	public void testGetBeans() throws ContextException
	{
		final String name1 = "bean1";
		final Object obj1 = new StringBuffer();
		final String name2 = "bean2";
		final Object obj2 = new Hashtable();
		
		ctx.registerBean(name1, obj1);
		ctx.registerBean(name2, obj2);
		
		assertEquals(Arrays.asList(new Object[]{obj1}),
				ctx.getBeans(StringBuffer.class));
	}
	
	public void testGetBean() throws ContextException
	{
		final String name1 = "bean1";
		final Object obj1 = new StringBuffer();
		final String name2 = "bean2";
		final Object obj2 = new Hashtable();
		
		ctx.registerBean(name1, obj1);
		ctx.registerBean(name2, obj2);
		
		assertEquals(obj2, ctx.getBean(Hashtable.class));
	}
	
	@Test(expectedExceptions = ContextException.class)
	public void testGetBean_Multiple() throws ContextException
	{
		final String name1 = "bean1";
		final Object obj1 = new Hashtable();
		final String name2 = "bean2";
		final Object obj2 = new Hashtable();
		
		ctx.registerBean(name1, obj1);
		ctx.registerBean(name2, obj2);
		
		ctx.getBean(Hashtable.class);
	}
	
	@Test(expectedExceptions = ContextException.class)
	public void testGetBean_NoMatch() throws ContextException
	{
		final String name1 = "bean1";
		final Object obj1 = new StringBuffer();
		final String name2 = "bean2";
		final Object obj2 = new Hashtable();
		
		ctx.registerBean(name1, obj1);
		ctx.registerBean(name2, obj2);
		
		ctx.getBean(Integer.class);
	}
	
	public void testGetProxy() throws ContextException
	{
		/* A default proxy is used initially */
		assertSame(ProxyFactory.getInstance(), ctx.getProxyFactory());
		
		IProxyFactory p = new ProxyFactory();
		ctx.registerBean("p", p);
		/* The new proxy bean replaces the original proxy */
		assertSame(p, ctx.getProxyFactory());
		
		/* The default factory remains the same */
		assertSame(ProxyFactory.getInstance(),
				Context.getDefaultProxyFactory());
	}
}
