/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.aurorasoftworks.signal.runtime.ui.mvc;


import static org.testng.AssertJUnit.assertEquals;

import java.util.ArrayList;
import java.util.List;

import org.mockito.InOrder;
import org.mockito.Mockito;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import com.aurorasoftworks.signal.runtime.core.context.ContextException;
import com.aurorasoftworks.signal.runtime.core.context.EagerInitBeanReference;
import com.aurorasoftworks.signal.runtime.core.context.IBeanReference;


@Test
public class AbstractDispatcherTest
{
	private AbstractMIDletDispatcher dispatcher;
	
	private TestContext ctx;
	
	@BeforeMethod
	protected void setUp()
	{
		TestContext.setDefaultProxyFactory(
				new MockControllerProxyFactory());
		ctx = new TestContext();
		dispatcher = createDispatcher();
		dispatcher.setContext(ctx);
	}
	
	protected AbstractMIDletDispatcher createDispatcher()
	{
		return new TestDispatcher();
	}
	
	public void testActivateDeactivate() throws Exception
	{
		ISimpleController ctl1 = Mockito.mock(ISimpleController.class);
		ISimpleController ctl2 = Mockito.mock(ISimpleController.class);
		
		IBeanReference ref1 = new EagerInitBeanReference("ctl1", ctl1);
		IBeanReference ref2 = new EagerInitBeanReference("ctl2", ctl2);
		
		ctx.processBean(ref1, dispatcher.postProcessBean(ref1));
		ctx.processBean(ref2, dispatcher.postProcessBean(ref2));
		
		ISimpleController ctlProxy1 = (ISimpleController) ctx.getBean(
				ref1.getName());
		ISimpleController ctlProxy2 = (ISimpleController) ctx.getBean(
				ref2.getName());
		
		ctlProxy1.perform();
		ctlProxy2.perform();
		
		InOrder inOrder = Mockito.inOrder(ctl1, ctl2);
		inOrder.verify(ctl1).activate();
		inOrder.verify(ctl1).perform();
		inOrder.verify(ctl1).deactivate();
		inOrder.verify(ctl2).activate();
		inOrder.verify(ctl2).perform();
		
		Mockito.verifyNoMoreInteractions(ctl1, ctl2);
	}
	
	@Test(dependsOnMethods = "testActivateDeactivate")
	public void testDispatcherAware() throws Exception
	{
		ISimpleDispatcherAwareController ctl1 = Mockito.mock(
				ISimpleDispatcherAwareController.class);
		ISimpleDispatcherAwareController ctl2 = Mockito.mock(
				ISimpleDispatcherAwareController.class);
		
		IBeanReference ref1 = new EagerInitBeanReference("ctl1", ctl1);
		IBeanReference ref2 = new EagerInitBeanReference("ctl2", ctl2);
		
		ctx.processBean(ref1, dispatcher.postProcessBean(ref1));
		ctx.processBean(ref2, dispatcher.postProcessBean(ref2));
		
		ISimpleController ctlProxy1 = (ISimpleDispatcherAwareController) ctx.getBean(
				ref1.getName());
		ISimpleController ctlProxy2 = (ISimpleDispatcherAwareController) ctx.getBean(
				ref2.getName());
		
		Mockito.when(ctl1.getDispatcher()).thenReturn(null);
		Mockito.when(ctl2.getDispatcher()).thenReturn(null);
		ctlProxy1.perform();
		ctlProxy2.perform();
		
		Mockito.when(ctl1.getDispatcher()).thenReturn(dispatcher);
		ctlProxy1.perform();
		
		InOrder inOrder = Mockito.inOrder(ctl1, ctl2);
		/* The first invocation returns null */
		inOrder.verify(ctl1).getDispatcher();
		inOrder.verify(ctl1).setDispatcher(dispatcher);
		inOrder.verify(ctl1).activate();
		inOrder.verify(ctl1).perform();
		inOrder.verify(ctl1).deactivate();
		inOrder.verify(ctl2).getDispatcher();
		inOrder.verify(ctl2).setDispatcher(dispatcher);
		inOrder.verify(ctl2).activate();
		inOrder.verify(ctl2).perform();
		inOrder.verify(ctl2).deactivate();
		/* This time the setter should not be invoked */
		inOrder.verify(ctl1).getDispatcher();
		inOrder.verify(ctl1).activate();
		inOrder.verify(ctl1).perform();
		
		Mockito.verifyNoMoreInteractions(ctl1, ctl2);
	}
	
	@Test(dependsOnMethods = "testActivateDeactivate")
	public void testFlow() throws Exception
	{
		final List<Object> flowCaller = new ArrayList<Object>();
		ISimpleController ctl = Mockito.mock(ISimpleController.class);
		ISimpleFlowController flowCtl = Mockito.mock(
				ISimpleFlowController.class);
		ISimpleFlowController flowCtlInterceptor
				= new SimpleFlowControllerInterceptor(flowCtl)
				{
					public void setCaller(IController caller)
					{
						flowCaller.add(caller);
					}
				};
		
		IBeanReference ref1 = new EagerInitBeanReference("ctl1", ctl);
		IBeanReference ref2 = new EagerInitBeanReference("ctl2",
				flowCtlInterceptor);
		
		ctx.processBean(ref1, dispatcher.postProcessBean(ref1));
		ctx.processBean(ref2, dispatcher.postProcessBean(ref2));
		
		ISimpleController ctlProxy = (ISimpleController) ctx.getBean(
				ref1.getName());
		ISimpleFlowController flowCtlProxy = (ISimpleFlowController)
				ctx.getBean(ref2.getName());
		
		assertEquals(0, dispatcher.getFlowLevel());
		ctlProxy.perform();
		assertEquals(0, dispatcher.getFlowLevel());
		assertEquals(0, flowCaller.size());
		
		flowCtlProxy.perform();
		assertEquals(1, dispatcher.getFlowLevel());
		assertEquals(1, flowCaller.size());
		
		IReturnEvent firstCaller = (IReturnEvent) flowCaller.get(0);
		flowCaller.clear();
		
		flowCtlProxy.perform();
		assertEquals(2, dispatcher.getFlowLevel());
		assertEquals(1, flowCaller.size());
		IReturnEvent secondCaller = (IReturnEvent) flowCaller.get(0);

		secondCaller.onReturn();
		assertEquals(1, dispatcher.getFlowLevel());
		
		firstCaller.onReturn();
		assertEquals(0, dispatcher.getFlowLevel());
	}
	
	@Test(expectedExceptions = Exception.class)
	public void testBeginFlowException() throws ContextException, Exception
	{
		ISimpleFlowController flowCtl = Mockito.mock(
				ISimpleFlowController.class);
		IBeanReference ref = new EagerInitBeanReference("ctl",
				flowCtl);
		
		ctx.processBean(ref, dispatcher.postProcessBean(ref));
		ISimpleFlowController flowCtlProxy = (ISimpleFlowController)
				ctx.getBean(ref.getName());
		
		assertEquals(0, dispatcher.getFlowLevel());
		flowCtlProxy.perform();
	}
}
