/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.aurorasoftworks.signal.runtime.core.context.proxy;


import java.util.Hashtable;


/**
 * Default implementation of the {@link IProxyFactory} interface.
 *
 * @author Marek
 */
public class ProxyFactory implements IProxyFactory
{
	private static final IProxyFactory INSTANCE = new ProxyFactory();

	public static final String PROXY_CLASS_SUFFIX = "__Proxy";
	
	public static final String PROXY_META_CLASS_SUFFIX = "__ProxyClass";
	
	public static IProxyFactory getInstance()
	{
		return INSTANCE;
	}
	
	public static String getProxyClassName(Class proxyClass)
	{
		return getBaseName(proxyClass.getName()) + PROXY_CLASS_SUFFIX;
	}
	
	public static String getProxyClassName(IProxyTarget proxyTarget)
	{
		return getProxyClassName(proxyTarget.getClass());
	}
	
	public static String getProxyMetaClassName(Class proxyClass)
	{
		return getBaseName(proxyClass.getName()) + PROXY_META_CLASS_SUFFIX;
	}
	
	private static String getBaseName(String name)
	{
		String result;
		
		if (name.endsWith(PROXY_CLASS_SUFFIX))
		{
			result = name.substring(
					0, name.length() - PROXY_CLASS_SUFFIX.length());
		}
		else
		{
			result = name;
		}
		
		return result;
	}

	public static String getProxyMetaClassName(IProxyTarget proxyTarget)
	{
		return getProxyMetaClassName(proxyTarget.getClass());
	}
	
	private Hashtable proxyClasses = new Hashtable();

	/**
	 * {@inheritDoc}
	 */
	public IProxy createProxy(IProxyTarget target, IInvocationHandler handler)
			throws Exception
	{
		AbstractProxy proxy = doCreateProxy(target);
		proxy.setHandler(handler);
		return proxy;
	}

	protected AbstractProxy doCreateProxy(
			IProxyTarget target) throws Exception
	{
		AbstractProxy proxy = doCreateProxy(target.getClass());
		proxy.setTarget(target);
		return proxy;
	}
	
	protected AbstractProxy doCreateProxy(
			Class targetClass) throws Exception
	{
		Class proxyClass = AbstractProxy.class.isAssignableFrom(targetClass)
				? targetClass : loadClass(getProxyClassName(targetClass));

		AbstractProxy proxy = (AbstractProxy) proxyClass.newInstance();
		return proxy;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public IProxy createLazyInitProxy(IBeanDelegate reference)
			throws Exception
	{
		AbstractProxy proxy = doCreateProxy(reference.getBeanClass());
		IMethodInterceptor interceptor = new LazyInitMethodInterceptor(
				reference);
		IInvocationHandler handler = getProxyClass(reference.getBeanClass());
		
		proxy.setHandler(interceptor);
		interceptor.setNextHandler(handler);
		
		return proxy;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public IProxy createProxy(IProxyTarget target,
			IMethodInterceptor [] interceptors)	throws Exception
	{
		AbstractProxy proxy = doCreateProxy(target);
		IInvocationHandler handler = getProxyClass(
				target.getClass());
		
		if ((interceptors != null) && (interceptors.length > 0))
		{
			proxy.setHandler(interceptors[0]);
			
			for (int i = 0; i < interceptors.length - 1; i++)
			{
				interceptors[i].setNextHandler(interceptors[i + 1]);
			}
			
			interceptors[interceptors.length - 1].setNextHandler(handler);
		}
		else
		{
			proxy.setHandler(handler);
		}
		
		return proxy;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public synchronized IProxyClass getProxyClass(Class targetClass)
			throws Exception
	{
		IProxyClass result = (IProxyClass) proxyClasses.get(targetClass);
		
		if (result == null)
		{
			result = (IProxyClass) loadClass(
					getProxyMetaClassName(targetClass)).newInstance();
			proxyClasses.put(targetClass, result);
		}
		
		return result;
	}

	protected Class loadClass(String className) throws Exception
	{
		return Class.forName(className);
	}
}
