/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.aurorasoftworks.signal.runtime.core.context.proxy;


import static org.testng.AssertJUnit.assertEquals;
import static org.testng.AssertJUnit.assertNotNull;
import static org.testng.AssertJUnit.assertSame;
import static org.testng.AssertJUnit.assertTrue;

import org.mockito.Mockito;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import com.aurorasoftworks.signal.runtime.core.context.proxy.AbstractMethodHandler;
import com.aurorasoftworks.signal.runtime.core.context.proxy.AbstractMethodInterceptor;
import com.aurorasoftworks.signal.runtime.core.context.proxy.AbstractProxy;
import com.aurorasoftworks.signal.runtime.core.context.proxy.AbstractProxyClass;
import com.aurorasoftworks.signal.runtime.core.context.proxy.IInvocationHandler;
import com.aurorasoftworks.signal.runtime.core.context.proxy.IMethodHandler;
import com.aurorasoftworks.signal.runtime.core.context.proxy.IMethodInterceptor;
import com.aurorasoftworks.signal.runtime.core.context.proxy.IProxyClass;
import com.aurorasoftworks.signal.runtime.core.context.proxy.IProxyFactory;
import com.aurorasoftworks.signal.runtime.core.context.proxy.IProxyTarget;
import com.aurorasoftworks.signal.runtime.core.context.proxy.MethodNotFoundException;
import com.aurorasoftworks.signal.runtime.core.context.proxy.ProxyFactory;


@Test
public class ProxyFactoryTest
{
	static class NoProxy implements IProxyTarget
	{
	}
	
	interface IFormatter
	{
		String format(int i);
	}
	
	static class Formatter implements IProxyTarget, IFormatter
	{
		private static int instanceCount;
		
		public synchronized static int getInstanceCount()
		{
			return instanceCount;
		}

		public synchronized static void resetInstanceCount()
		{
			Formatter.instanceCount = 0;
		}
		
		public Formatter()
		{
			synchronized (Formatter.class)
			{
				instanceCount++;
			}
		}
		
		public String format(int i)
		{
			return String.valueOf(i);
		}
	}
	
	/**
	 * Implementations of {@link AbstractProxy} are normally
	 * generated automatically. Since the code generator is implemented
	 * in a different subproject, however, this manually coded
	 * class is used for unit testing.
	 * 
	 * Tests of generated classes are implemented in
	 * JavaSourceContextGeneratorTest.
	 */
	static class Formatter__Proxy extends AbstractProxy
			implements IFormatter
	{
		private static final IProxyClass PROXY_CLASS
				= new Formatter__ProxyClass();
		
		private static final IMethodHandler FORMAT_HANDLER
				= PROXY_CLASS.getMethodHandler(
						"format", new Class[]{int.class});
		
		public String format(int i)
		{
			try
			{
				return (String) handler.invoke(target,
						FORMAT_HANDLER, new Integer[]{i});
			}
			catch (Exception e)
			{
				throw new RuntimeException(e.toString());
			}
		}	
	}
	
	/**
	 * Implementations of {@link AbstractProxyClass} are normally
	 * generated automatically. Since the code generator is implemented
	 * in a different subproject, however, this manually coded
	 * class is used for unit testing.
	 * 
	 * Tests of generated classes are implemented in
	 * JavaSourceContextGeneratorTest.
	 */
	static class Formatter__ProxyClass extends AbstractProxyClass
	{
		static IMethodHandler formatHandler = new AbstractMethodHandler(
				"format", new Class[]{int.class})
		{	
			public Object invoke(IProxyTarget target, Object[] args)
					throws Exception
			{
				return ((IFormatter) target).format(
						((Integer) args[0]).intValue());
			}
		};
		
		Formatter__ProxyClass()
		{
			registerHandlers(new IMethodHandler[]{formatHandler});
		}
	}
	
	static class PrependingMethodInterceptor extends AbstractMethodInterceptor
	{
		private String textToPrepend;
		
		public PrependingMethodInterceptor(String textToPrepend)
		{
			this.textToPrepend = textToPrepend;
		}

		public Object invoke(IProxyTarget target, IMethodHandler handler,
				Object[] args) throws Exception
		{
			return textToPrepend + nextHandler.invoke(
					target, handler, args);
		}	
	}

	private IProxyFactory factory;
	
	@BeforeMethod
	protected void setUp()
	{
		factory = new ProxyFactory();
	}

	public void testGetProxyClass() throws Exception
	{
		assertTrue(factory.getProxyClass(Formatter.class)
				instanceof Formatter__ProxyClass);
		
		/* Make sure that proxy class cache works */
		assertSame(factory.getProxyClass(Formatter.class),
				factory.getProxyClass(Formatter.class));
		
		assertEquals("format", factory.getProxyClass(Formatter.class)
				.getMethodHandler("format", new Class[]{int.class})
				.getMethodName());
	}
	
	@Test(dependsOnMethods = "testGetProxyClass",
			expectedExceptions = MethodNotFoundException.class)
	public void testGetProxyClass_GetMethodHandler() throws Exception
	{
		factory.getProxyClass(Formatter.class)
				.getMethodHandler("formatInt", new Class[]{int.class});
	}
	
	@Test(dependsOnMethods = "testGetProxyClass",
			expectedExceptions = MethodNotFoundException.class)
	public void testGetProxyClass_GetMethodHandler2() throws Exception
	{
		factory.getProxyClass(Formatter.class)
				.getMethodHandler("format", new Class[]{int.class, float.class});
	}
	
	@Test(dependsOnMethods = "testGetProxyClass",
			expectedExceptions = MethodNotFoundException.class)
	public void testGetProxyClass_GetMethodHandler3() throws Exception
	{
		factory.getProxyClass(Formatter.class)
				.getMethodHandler("format", null);
	}
	
	@Test(dependsOnMethods = "testGetProxyClass",
			expectedExceptions = MethodNotFoundException.class)
	public void testGetProxyClass_GetMethodHandler4() throws Exception
	{
		factory.getProxyClass(Formatter.class)
				.getMethodHandler("format", new Class[]{double.class});
	}

	@Test(dependsOnMethods = "testGetProxyClass",
			expectedExceptions = Exception.class)
	public void testGetProxyClass_MissingProxy() throws Exception
	{
		factory.getProxyClass(NoProxy.class);
	}
	
	@Test(dependsOnMethods = "testGetProxyClass")
	public void testCreateProxy_WithHandler() throws Exception
	{
		IInvocationHandler handler = Mockito.mock(
				IInvocationHandler.class);
		Formatter formatter = new Formatter();
		IFormatter proxy = (IFormatter) factory.createProxy(
				formatter, handler);
		assertNotNull(proxy);
		Mockito.when(proxy.format(10)).thenReturn("10");
		proxy.format(10);
		
		Mockito.verify(handler).invoke(formatter,
				factory.getProxyClass(Formatter.class).getMethodHandler(
						"format", new Class[]{int.class}),
				new Object[]{10});
	}
	
	@Test(dependsOnMethods = "testCreateProxy_WithHandler")
	public void testCreateProxy_WithTarget() throws Exception
	{
		Formatter formatter = new Formatter();
		IFormatter proxy = (IFormatter) factory.createProxy(formatter,
				new IMethodInterceptor[]{});
		assertEquals("20", proxy.format(20));
		
		proxy = (IFormatter) factory.createProxy(formatter,
				new IMethodInterceptor[]
				{
					new PrependingMethodInterceptor("a"),
					new PrependingMethodInterceptor("b"),
				});
		assertEquals("ab30", proxy.format(30));
	}
	
	/*@Test(dependsOnMethods = "testCreateProxy_WithTarget")
	public void testCreateLazyInitProxy() throws Exception
	{
		Formatter.resetInstanceCount();
		assertEquals(0, Formatter.getInstanceCount());
		
		IFormatter proxy = (IFormatter) factory.createLazyInitProxy(
				new LazyInitBeanReference("name", Formatter.class)
		{
			protected Object createBean() throws Exception
			{
				return new Formatter();
			}
		});
		assertEquals(0, Formatter.getInstanceCount());
		
		assertEquals("2", proxy.format(2));
		assertEquals(1, Formatter.getInstanceCount());
	}*/
	
	@Test(expectedExceptions = Exception.class)
	public void testCreateProxy_MissingProxy() throws Exception
	{
		factory.createProxy(new NoProxy(),
				Mockito.mock(IInvocationHandler.class));
	}
}
