/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.aurorasoftworks.signal.runtime.core.context.proxy;


/**
 * A base class for implementations of {@link IMethodHandler}
 * 
 * @author Marek
 */
public abstract class AbstractMethodHandler implements IMethodHandler
{
	private String methodName;
	
	private Class [] argumentTypes;
	
	/**
	 * Creates a new AbstractMethodHandler object.
	 * 
	 * @param methodName     name of the method this objects represents
	 * @param argumentTypes  types of arguments of the method
	 *                       this object represents
	 */
	protected AbstractMethodHandler(
			String methodName, Class [] argumentTypes)
	{
		this.methodName = methodName;
		this.argumentTypes = argumentTypes;
	}
	
	/**
	 * Creates a new AbstractMethodHandler object for a method
	 * that accepts a single argument.
	 * 
	 * @param methodName     name of the method this objects represents
	 * @param argumentType   type of the single argument of the method
	 *                       this object represents
	 */
	protected AbstractMethodHandler(
			String methodName, Class argumentType)
	{
		this(methodName, new Class[]{argumentType});
	}
	
	/**
	 * Creates a new AbstractMethodHandler object for a method
	 * that does not accept arguments.
	 * 
	 * @param methodName     name of the method this objects represents
	 */
	protected AbstractMethodHandler(String methodName)
	{
		this(methodName, new Class[]{});
	}
	
	/**
	 * {@inheritDoc}
	 */
	public boolean matchesArgumentTypes(Class [] types)
	{
		boolean result;
		
		if ((types != null) && (types.length == argumentTypes.length))
		{
			result = true;
			
			for (int i = 0; i < argumentTypes.length; i++)
			{
				if (!types[i].equals(argumentTypes[i]))
				{
					result = false;
					break;
				}
			}
		}
		else
		{
			result = false;
		}
		
		return result;
	}

	/**
	 * {@inheritDoc}
	 */
	public String getMethodName()
	{
		return methodName;
	}

	/**
	 * {@inheritDoc}
	 */
	public Class [] getArgumentTypes()
	{
		return argumentTypes;
	}
}
