/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.aurorasoftworks.signal.runtime.core.context.proxy;


import java.lang.reflect.Proxy;


/**
 * A factory for AOP proxies. Proxies can only be created for
 * classes that implement {@link IProxyTarget} and are used in
 * a context file that is processed by a context generator.
 * 
 * @author Marek
 */
public interface IProxyFactory
{
	/**
	 * Creates a new proxy for the specified object.
	 * All method calls are forwarded to the provided
	 * handler. Depending on how the handler is implemented
	 * calls may or may not be invoked on the target
	 * object.
	 * 
	 * This method is similar to
	 * {@link Proxy#newProxyInstance(ClassLoader, Class[], java.lang.reflect.InvocationHandler)}.
	 * 
	 * @param target  an object to create a proxy for
	 * @param handler an invocation handler to forward
	 *                intercepted method calls to
	 * @return value to be returned by the intercepted call
	 * @throws Exception
	 */
	IProxy createProxy(IProxyTarget target, IInvocationHandler handler)
			throws Exception;
	
	/**
	 * Creates a new proxy for the specified object using the specified
	 * interceptors. This method creates a sequence of method invocation
	 * handlers that contains the specified interceptors and the target
	 * object at the end. Each handler in the chain can choose whether
	 * to forward the call to the next handler or not.
	 * 
	 * @param target       an object to create a proxy for
	 * @param interceptors an array of method interceptors
	 * @return 
	 * @throws Exception
	 */
	IProxy createProxy(IProxyTarget target,
			IMethodInterceptor [] interceptors)	throws Exception;
	
	/**
	 * Creates a lazy init proxy for the specified bean reference.
	 * When any method is invoked on the returned proxy the bean
	 * wrapped by the reference is created by calling
	 * {@link IBeanReference#getBean()}.
	 * 
	 * @param reference a bean reference to be wrapped by a lazy init proxy
	 * @return a lazy init proxy
	 * @throws Exception
	 */
	IProxy createLazyInitProxy(IBeanDelegate reference)
			throws Exception;
	
	/**
	 * Returns an {@link IProxyClass} object for the specified class.
	 * @param targetClass a class to return a proxy class for
	 * @return proxy class
	 * @throws Exception
	 */
	IProxyClass getProxyClass(
			Class targetClass) throws Exception;
}
