/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.aurorasoftworks.signal.tools.core.context.generator.source.java;


import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.formatCast;
import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.formatLiteral;
import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.formatWrapPrimitive;
import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.isPrimitive;
import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.isPrimitiveWrapper;
import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.wrapperToPrimitive;

import java.io.File;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import com.aurorasoftworks.signal.runtime.core.context.Context;
import com.aurorasoftworks.signal.tools.core.context.generator.ContextGeneratorException;
import com.aurorasoftworks.signal.tools.core.context.generator.source.AbstractContextSourceGenerator;
import com.aurorasoftworks.signal.tools.core.context.loader.IBeanInfo;
import com.aurorasoftworks.signal.tools.core.context.loader.IBeanReference;
import com.aurorasoftworks.signal.tools.core.context.loader.IContextReference;


public abstract class AbstractJavaSourceContextGenerator<T> extends
		AbstractContextSourceGenerator<T>
{
	protected static final String JAVA_EXTENSION = ".java";

	protected String contextClassName;
	
	protected String contextPackageName;
	
	protected String contextBaseClassFullName = Context.class.getName();
	
	protected void updateContextClassName(String fullContextClassName)
	{
		this.contextClassName
				= com.aurorasoftworks.signal.tools.core.util.BeanUtils
						.getClassName(fullContextClassName);
		this.contextPackageName
				= com.aurorasoftworks.signal.tools.core.util.BeanUtils
						.getPackageName(fullContextClassName);
	}

	protected String formatGetBean(String name)
	{
		return formatThis() + ".getBean("
				+ formatLiteral(name) + ")";
	}

	protected String formatGetBean(String name, Class<?> cls)
	{
		return formatCast(formatGetBean(name), cls);
	}
	
	protected String formatGetBean(IBeanInfo info)
	{
		return formatGetBean(info.getName(), info.getBeanClass());
	}

	protected String formatGetBeanReference(String name)
	{
		return formatCast(formatThis() + ".getBeanReference("
				+ formatLiteral(name) + ")",
				com.aurorasoftworks.signal.runtime.core.context.IBeanReference.class);
	}

	protected String formatThis()
	{
		return contextClassName + ".this";
	}

	protected String formatArgs(Object [] args)
	{
		Class<?> [] classes = null;
		
		if (args != null)
		{
			classes = new Class<?>[args.length];
			for (int i = 0; i < args.length; i++)
			{
				classes[i] = args[i].getClass();
			}
		}
		
		return formatArgs(args, classes);
	}

	protected String formatArgs(Object [] args, Class<?> [] classes)
	{
		StringBuilder builder = new StringBuilder();
		
		if (args != null)
		{
			for (int i = 0; i < args.length; i++)
			{
				builder.append(formatArg(args[i], classes[i]));
				if (i < args.length - 1)
				{
					builder.append(", ");
				}
			}
		}
		
		return builder.toString();
	}

	protected String formatArg(Object arg)
	{
		return formatArg(arg,
				(arg instanceof IBeanReference)
						? ((IBeanReference) arg).getObject().getClass()
						: arg.getClass());
	}

	protected String formatArg(Object arg, Class<?> targetClass)
	{
		String result;
		
		if (arg == null)
		{
			result = "null";
		}
		else if (arg instanceof IBeanReference)
		{
			result = formatGetBean(((IBeanReference) arg).getName(),
					targetClass);
		}
		else if (arg instanceof IContextReference)
		{
			result = formatThis();
		}
		else if (arg instanceof String)
		{
			result = formatLiteral((String) arg);
		}
		else if (isPrimitive(targetClass))
		{
			result = formatLiteral(arg);
		}
		else if (isPrimitiveWrapper(targetClass))
		{
			result = formatWrapPrimitive(formatLiteral(arg),
					wrapperToPrimitive(targetClass));
		}
		else if (arg instanceof List<?>)
		{
			result = formatList((List<?>) arg);
		}
		else if (arg instanceof Set<?>)
		{
			result = formatSet((Set<?>) arg);
		}
		else if (arg instanceof Map<?, ?>)
		{
			result = formatMap((Map<?, ?>) arg);
		}
		else
		{
			result = arg.toString();
		}
		
		return result;
	}

	protected String formatList(List<?> list)
	{
		return doFormatVector(list);
	}

	protected String doFormatVector(Collection<?> list)
	{
		StringBuilder buf = new StringBuilder("new java.util.Vector(){{");
		
		for (Object obj : list)
		{
			buf.append("addElement(");
			buf.append(formatArg(obj));
			buf.append("); ");
		}
		
		buf.append("}}");
		return buf.toString();
	}

	protected String formatSet(Set<?> set)
	{
		return doFormatVector(set);
	}

	protected String formatMap(Map<?, ?> map)
	{
		StringBuilder buf = new StringBuilder(
				"new java.util.Hashtable(){{");
		
		for (Entry<?, ?> entry : map.entrySet())
		{
			buf.append("put(");
			buf.append(formatArg(entry.getKey()));
			buf.append(", ");
			buf.append(formatArg(entry.getValue()));
			buf.append("); ");
		}
		
		buf.append("}}");
		return buf.toString();
	}

	@Override
	protected void onBeanGenerationStarted(IBeanInfo bean)
	{
		getWriter().println("/* Begin " + bean.getName() + " */");
	}

	@Override
	protected void onBeanGenerationCompleted(IBeanInfo bean)
			throws ContextGeneratorException
	{
		getWriter().println("/* End " + bean.getName() + " */\n");
	}

	protected File createOutputDirectory(File outputDirectory)
	{
		File dir = outputDirectory;
		
		if (contextPackageName != null)
		{
			dir = new File(dir, contextPackageName.replace(
					'.', File.separatorChar));
		}
		
		dir.mkdirs();
		
		return dir;
	}

	public String getContextBaseClassFullName()
	{
		return contextBaseClassFullName;
	}

	public void setContextBaseClassFullName(String contextBaseClassFullName)
	{
		this.contextBaseClassFullName = contextBaseClassFullName;
	}
}
