/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.aurorasoftworks.signal.tools.core.context.generator.source.java;


import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.formatCast;
import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.formatClass;
import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.formatLiteral;
import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.getGeneratedOnComment;

import java.beans.PropertyDescriptor;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;


import org.springframework.beans.BeanUtils;

import com.aurorasoftworks.signal.runtime.core.context.EagerInitBeanReference;
import com.aurorasoftworks.signal.runtime.core.context.LazyInitBeanReference;
import com.aurorasoftworks.signal.tools.core.context.ContextProcessingException;
import com.aurorasoftworks.signal.tools.core.context.generator.ContextGeneratorException;
import com.aurorasoftworks.signal.tools.core.context.loader.IBeanInfo;
import com.aurorasoftworks.signal.tools.core.context.loader.IContextLoader;
import com.aurorasoftworks.signal.tools.core.context.loader.IContextLoaderListener;


public class JavaSourceContextGenerator<T>
		extends AbstractJavaSourceContextGenerator<T>
		implements IContextLoaderListener
{
	abstract class State implements IContextLoaderListener
	{
		protected IBeanInfo info;
		
		State(IBeanInfo info)
		{
			this.info = info;
		}

		public void onBeanInstantiated(IBeanInfo info, Constructor<?> ctor,
				Object[] args) throws ContextGeneratorException
		{			
			doRegisterBean(info, "new "
					+ info.getBean().getClass().getName() + "("
					+ formatArgs(args, ctor.getParameterTypes()) + ")");
		}
		
		public void onBeanInstantiated(IBeanInfo info, Method factoryMethod,
				Object[] args) throws ContextGeneratorException
		{
			doRegisterBean(info,
					factoryMethod.getDeclaringClass().getName()
					+ "." + factoryMethod.getName() + "("
					+ formatArgs(args, factoryMethod.getParameterTypes()) + ")");	
		}

		public void onBeanInstantiated(IBeanInfo info,
				IBeanInfo factoryBeanInfo, Method factoryMethod,
				Object[] args) throws ContextGeneratorException
		{	
			doRegisterBean(info,
					formatGetBean(factoryBeanInfo)
					+ "." + factoryMethod.getName() + "("
					+ formatArgs(args, factoryMethod.getParameterTypes()) + ")");
		}
		
		public void onBeanPropertySet(IBeanInfo info, String propertyName,
				Object value) throws ContextGeneratorException
		{	
			PropertyDescriptor pd = BeanUtils.getPropertyDescriptor(
					info.getBean().getClass(), propertyName);
			getWriter().println(formatGetBean(info) 
					+ "." + pd.getWriteMethod().getName()
					+ "(" + formatArg(value, pd.getPropertyType()) + ");");
		}
		
		public void onBeanMethodInvoked(IBeanInfo info,
				Method method, Object [] args)
				throws ContextGeneratorException
		{			
			getWriter().println(formatGetBean(info) 
					+ "." + method.getName()
					+ "(" + formatArgs(args, method.getParameterTypes()) + ");");
		}
		
		public void onBeanPreProcess(IBeanInfo info,
				IBeanInfo processorInfo)
		{
			doProcessBean(info, processorInfo, "preProcessBean");
		}
		
		public void onBeanPostProcess(IBeanInfo info,
				IBeanInfo processorInfo)
		{
			doProcessBean(info, processorInfo, "postProcessBean");
		}

		protected void doProcessBean(IBeanInfo info, IBeanInfo processorInfo,
				String methodName)
		{
			getWriter().println(formatThis() + ".processBean("
					+ formatGetBeanReference(info.getName())
					+ ", " + formatGetBean(processorInfo)
					+ "." + methodName + "("
					+ formatGetBeanReference(info.getName())
					+ "));");
		}
		
		protected abstract void doRegisterBean(IBeanInfo info, String expression);

		protected String formatGetBean(IBeanInfo processorInfo)
		{
			return JavaSourceContextGenerator.this.formatGetBean(
					processorInfo);
		}
		
		public void exit() throws ContextGeneratorException
		{
		}
	}
	
	class EagerInitState extends State
	{
		EagerInitState(IBeanInfo info)
		{
			super(info);
		}
		
		protected void doRegisterBean(IBeanInfo info, String expression)
		{
			getWriter().println(formatThis() + ".registerBean("
					+ formatLiteral(info.getName()) + ", "
					+ expression + ");");
		}
	}
	
	class LazyInitState extends State
	{
		LazyInitState(IBeanInfo info)
		{
			super(info);
		}
		
		protected void doProcessBean(IBeanInfo info, IBeanInfo processorInfo,
				String methodName)
		{
			if (info.equals(this.info))
			{
				String eagerInitRef = "new "
						+ EagerInitBeanReference.class.getName()
						+ "(" + formatLiteral(info.getName()) + ", result)";
				
				getWriter().println("result = " + formatThis()
						+ ".processLazyBean("
						+ formatGetBean(processorInfo)
						+ "." + methodName + "("
						+ eagerInitRef + "));");
			}
			else
			{
				super.doProcessBean(info, processorInfo, methodName);
			}
		}
		
		protected String formatGetBean(IBeanInfo info)
		{
			return info.equals(this.info)
					? formatCast("result", info.getBeanClass())
					: JavaSourceContextGenerator.this.formatGetBean(info);
		}
		
		protected void doRegisterBean(IBeanInfo info, String expression)
		{
			String referencePrefix = "new " + LazyInitBeanReference.class.getName() + "("
					+ formatLiteral(info.getName()) + ", "
					+ formatClass(info.getBeanClass()) + ") {"
					+ "protected Object createBean() throws Exception {";
			if (info.isProxyTarget())
			{
				getWriter().println(formatThis() + ".registerBean("
						+ formatLiteral(info.getName()) + ", "
						+ formatThis()
						+ ".getProxyFactory().createLazyInitProxy(");
			}
			else
			{
				getWriter().println(formatThis() + ".registerBeanReference(");
			}
			getWriter().indent();
			
			getWriter().println(referencePrefix);
			getWriter().println("Object result = " + expression + ";");
		}
		
		public void exit() throws ContextGeneratorException
		{
			getWriter().println("return result;");
			getWriter().println("}}");
			
			getWriter().unindent();
			
			if (info.isProxyTarget())
			{
				getWriter().println("));");				
			}
			else
			{
				getWriter().println(");");
			}
		}
	}
	
	private State state;
	
	private boolean generateProxyClasses = true;

	public void generateContext(String contextLocation,
			String contextClassFullName,
			IContextLoader<T> loader,
			File outputDirectory) throws ContextProcessingException
	{
		updateContextClassName(contextClassFullName);
		
		try
		{
			File classDirectory = createOutputDirectory(outputDirectory);
			
			Writer writer = new BufferedWriter(new FileWriter(
					new File(classDirectory,
							contextClassName + JAVA_EXTENSION)));
			try
			{
				generateContext(contextLocation,
						loader, writer);
			}
			finally
			{
				writer.close();
			}
			
			if (generateProxyClasses)
			{
				new JavaSourceProxyGenerator().generateProxies(
						outputDirectory, getProcessedBeans());
			
				new JavaSourceProxyClassGenerator()
						.generateProxyClasses(outputDirectory,
								getProcessedBeans());
			}
		}
		catch (IOException e)
		{
			throw new ContextProcessingException(
					"I/O error while generating a context", e);
		}
	}
	
	protected void generateContext(String contextLocation,
			String contextClassFullName,
			IContextLoader<T> loader, Writer writer)
			throws ContextProcessingException
	{
		updateContextClassName(contextClassFullName);

		super.generateContext(contextLocation, loader, writer);
	}
	
	@Override
	protected void doGenerateContext(String contextLocation,
			IContextLoader<T> loader)
			throws ContextProcessingException
	{
		if (contextPackageName != null)
		{
			getWriter().println("package " + contextPackageName + ";\n");
			getWriter().println();
		}
		
		getWriter().println(getGeneratedOnComment());
		getWriter().println("public class " + contextClassName
				+ " extends " + contextBaseClassFullName);
		getWriter().println("{");
		getWriter().indent();
		getWriter().println("public " + contextClassName
				+ "() throws " + Exception.class.getName());
		getWriter().println("{");
		getWriter().indent();
		
		super.doGenerateContext(contextLocation, loader);
		
		getWriter().unindent();
		getWriter().println("}");
		getWriter().unindent();
		getWriter().println("}");
		getWriter().println();
	}
	
	@Override
	protected IContextLoaderListener getContextLoaderListener()
	{
		return this;
	}

	public void onBeanInstantiated(IBeanInfo info, Constructor<?> ctor,
			Object[] args) throws ContextGeneratorException
	{
		checkBean(info);
		state.onBeanInstantiated(info, ctor, args);
	}

	public void onBeanInstantiated(IBeanInfo info, IBeanInfo factoryBeanInfo,
			Method factoryMethod, Object[] args)
			throws ContextGeneratorException
	{
		checkBean(info);
		state.onBeanInstantiated(info, factoryBeanInfo, factoryMethod, args);
	}

	public void onBeanInstantiated(IBeanInfo info, Method factoryMethod,
			Object[] args) throws ContextGeneratorException
	{
		checkBean(info);
		state.onBeanInstantiated(info, factoryMethod, args);
	}

	public void onBeanMethodInvoked(IBeanInfo info, Method method, Object[] args)
			throws ContextGeneratorException
	{
		checkBean(info);
		state.onBeanMethodInvoked(info, method, args);
	}
	
	public void onBeanPropertySet(IBeanInfo info, String propertyName,
			Object value) throws ContextGeneratorException
	{
		checkBean(info);
		state.onBeanPropertySet(info, propertyName, value);
	}

	public void onBeanPreProcess(IBeanInfo info, IBeanInfo processorInfo)
			throws ContextGeneratorException
	{
		checkBean(info);
		state.onBeanPreProcess(info, processorInfo);
	}
	
	public void onBeanPostProcess(IBeanInfo info, IBeanInfo processorInfo)
			throws ContextGeneratorException
	{
		checkBean(info);
		state.onBeanPostProcess(info, processorInfo);
	}

	@Override
	protected void onBeanGenerationCompleted(IBeanInfo bean)
			throws ContextGeneratorException
	{
		state.exit();
		super.onBeanGenerationCompleted(bean);
	}

	@Override
	protected void onBeanGenerationStarted(IBeanInfo bean)
	{
		super.onBeanGenerationStarted(bean);
		state = bean.isLazyInit()
				? new LazyInitState(bean) : new EagerInitState(bean);
	}

	public boolean getGenerateProxyClasses()
	{
		return generateProxyClasses;
	}

	public void setGenerateProxyClasses(boolean generateProxyClasses)
	{
		this.generateProxyClasses = generateProxyClasses;
	}
}
