package com.aurorasoftworks.signal.tools.core.context.generator.source.java;


import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.formatCast;
import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.formatClass;
import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.formatClasses;
import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.formatHandlerName;
import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.formatLiteral;
import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.formatUnwrapPrimitive;
import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.formatWrapPrimitive;
import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.getGeneratedOnComment;
import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.isPrimitive;
import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.primitiveToWrapper;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.aurorasoftworks.signal.runtime.core.context.proxy.AbstractMethodHandler;
import com.aurorasoftworks.signal.runtime.core.context.proxy.AbstractProxyClass;
import com.aurorasoftworks.signal.runtime.core.context.proxy.IProxyTarget;
import com.aurorasoftworks.signal.runtime.core.context.proxy.ProxyFactory;
import com.aurorasoftworks.signal.tools.core.context.generator.ContextGeneratorException;
import com.aurorasoftworks.signal.tools.core.context.generator.source.SourceWriter;
import com.aurorasoftworks.signal.tools.core.context.loader.IBeanInfo;
import com.aurorasoftworks.signal.tools.core.util.BeanUtils;


public class JavaSourceProxyClassGenerator
{
	private SourceWriter writer;
	
	public void generateProxyClasses(File outputDir,
			List<IBeanInfo> beans) throws ContextGeneratorException
	{
		try
		{
			for (IBeanInfo bean : beans)
			{
				if (bean.isProxyTarget())
				{
					generateProxyClass(outputDir, bean);
				}	
			}
		}
		catch (ContextGeneratorException e)
		{
			throw e;
		}
		catch (Exception e)
		{
			throw new ContextGeneratorException(e);
		}
	}
	
	protected void generateProxyClass(File outputDir, IBeanInfo bean)
			throws IOException, ContextGeneratorException
	{
		File proxyFile = new File(outputDir,
				(bean.getFullClassName().replace(
					'.', File.separatorChar)
					+ ProxyFactory.PROXY_META_CLASS_SUFFIX)
					+ AbstractJavaSourceContextGenerator.JAVA_EXTENSION);
		proxyFile.getParentFile().mkdirs();

		writer = new SourceWriter(new FileWriter(proxyFile));
		
		try
		{
			generateProxyClass(bean);
		}
		finally
		{
			writer.close();
			writer = null;
		}
	}
		
	protected void generateProxyClass(IBeanInfo bean) throws IOException,
			ContextGeneratorException
	{
		if (bean.getPackageName() != null)
		{
			writer.println("package " + bean.getPackageName() + ";");
			writer.println();
		}

		writer.println(getGeneratedOnComment());
		writer.println("public class " + BeanUtils.getClassName(
						ProxyFactory.getProxyMetaClassName(bean.getBeanClass()))
				+ " extends " + AbstractProxyClass.class.getName());
		writer.println("{");
		writer.indent();
		
		Map<String, List<String>> handlers
				= generateMethodInvocationHandlers(bean);
		
		generateCtor(bean, handlers);
		
		writer.unindent();
		writer.println("}");
		writer.println();
	}

	private void generateCtor(
			IBeanInfo bean, Map<String, List<String>> handlers)
	{
		writer.println("public " + BeanUtils.getClassName(
				ProxyFactory.getProxyMetaClassName(bean.getBeanClass()))
				+ "()");
		writer.println("{");
		writer.indent();
		
		for (Map.Entry<String, List<String>> e : handlers.entrySet())
		{
			if (e.getValue().size() == 1)
			{
				writer.println("registerHandler(" 
						+ e.getValue().get(0) + ");");
			}
			else
			{
				writer.println("registerHandlers(" 
						+ formatHandlerArray(e.getValue()) + ");");
			}
		}
		
		writer.unindent();
		writer.println("}");
		writer.println();
	}

	protected String formatHandlerArray(List<String> handlers)
	{
		StringBuilder buf = new StringBuilder("new "
				+ AbstractMethodHandler.class.getName() + "[]{");
		
		for (String h : handlers)
		{
			buf.append(h);
			buf.append(", ");
		}
		
		buf.append("}");
		
		return buf.toString();
	}

	protected Map<String, List<String>> generateMethodInvocationHandlers(
			IBeanInfo bean) throws ContextGeneratorException
	{
		Map<String, List<String>> handlerNames
				= new HashMap<String, List<String>>()
				{
					public List<String> get(Object key)
					{
						List<String> result = super.get(key);
						if (result == null)
						{
							result = new ArrayList<String>();
							put((String) key, result);
						}
						return result;
					}
				};
		
		Map<Method, Class<?>> methods = getInterfaceMethodMapping(bean);
		
		int i = 0;
		for (Map.Entry<Method, Class<?>> e : methods.entrySet())
		{
			String handlerName = generateMethodInvocationHandler(bean, i++,
					e.getKey(), e.getValue());
			handlerNames.get(e.getKey().getName()).add(handlerName);
		}
		
		return handlerNames;
	}

	protected String generateMethodInvocationHandler(
			IBeanInfo bean, int index, Method m, Class<?> intface)
	{
		String handlerName = formatHandlerName(index);
		
		writer.println("private static final "
				+ AbstractMethodHandler.class.getName() + " "
				+ handlerName + " = "
				+ formatMethodHandlerCtor(m));
		writer.println("{");
		writer.indent();
		
		generateMethodInvokerInvoke(bean, m, intface);
		
		writer.unindent();
		writer.println("};");
		writer.println();
		
		return handlerName;
	}
	
	protected String formatMethodHandlerCtor(Method m)
	{
		return "new " + AbstractMethodHandler.class.getName() + "("
				+ formatLiteral(m.getName())
				+ (m.getParameterTypes().length > 0
						?  (", " + formatMethodHandlerClasses(
								m.getParameterTypes()))
						: "")
				+ ")";
	}

	protected String formatMethodHandlerClasses(Class [] classes)
	{
		String result;
		
		if (classes.length == 1)
		{
			result = formatClass(classes[0]);
		}
		else
		{
			result = formatClasses(classes);
		}
		
		return result;
	}

	protected void generateMethodInvokerInvoke(
			IBeanInfo bean, Method m, Class<?> intface)
	{
		writer.println("public Object invoke("
				+ IProxyTarget.class.getName()
				+ " target, Object [] args) throws Exception");
		writer.println("{");
		writer.indent();
		
		String invocation = formatCast("target", intface) + "."
				+ m.getName()+ "(" + formatArguments(m) + ")";
		
		if (void.class.equals(m.getReturnType()))
		{
			writer.println(invocation + ";");
			writer.println("return null;");
		}
		else
		{
			writer.println("return "
					+ formatReturnValue(invocation, m) + ";");
		}
		
		writer.unindent();
		writer.println("}");
	}

	protected String formatReturnValue(String expression, Method m)
	{
		String result;
		
		if (isPrimitive(m.getReturnType()))
		{
			result = formatWrapPrimitive(expression, m.getReturnType());
		}
		else
		{
			result  = expression;
		}
		
		return result;
	}

	protected String formatArguments(Method m)
	{
		StringBuilder buf = new StringBuilder();
		
		Class<?> [] paramTypes = m.getParameterTypes();
		
		for (int i = 0; i < paramTypes.length; i++)
		{
			buf.append(formatArgument(i, paramTypes[i]));
			if (i < paramTypes.length - 1)
			{
				buf.append(", ");
			}
		}
		
		return buf.toString();
	}

	protected String formatArgument(int i, Class<?> cls)
	{
		String result;
		String arg = "args[" + i + "]";
		
		if (isPrimitive(cls))
		{
			result = formatUnwrapPrimitive(
					formatCast(arg, primitiveToWrapper(cls)), cls);
		}
		else
		{
			result = formatCast(arg, cls);
		}
		
		return result;
	}

	protected Map<Method, Class<?>> getInterfaceMethodMapping(
			IBeanInfo bean)	throws ContextGeneratorException
	{
		try
		{
			return BeanUtils.getInterfaceMethodMapping(
					bean.getBeanClass());
		} catch (Exception e)
		{
			throw new ContextGeneratorException(e);
		}
	}
}
