package com.aurorasoftworks.signal.tools.core.context.generator.source.java;


import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.formatClasses;
import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.formatHandlerName;
import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.formatLiteral;
import static com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceUtils.getGeneratedOnComment;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import com.aurorasoftworks.signal.runtime.core.context.proxy.AbstractProxy;
import com.aurorasoftworks.signal.runtime.core.context.proxy.IMethodHandler;
import com.aurorasoftworks.signal.runtime.core.context.proxy.IProxyClass;
import com.aurorasoftworks.signal.runtime.core.context.proxy.ProxyFactory;
import com.aurorasoftworks.signal.tools.core.context.generator.ContextGeneratorException;
import com.aurorasoftworks.signal.tools.core.context.generator.source.SourceWriter;
import com.aurorasoftworks.signal.tools.core.context.loader.IBeanInfo;
import com.aurorasoftworks.signal.tools.core.util.BeanUtils;



public class JavaSourceProxyGenerator extends AbstractJavaSourceProxyGenerator
{
	protected static final String PROXY_CLASS_NAME = "_PROXY_CLASS";
	
	private SourceWriter writer;
	
	public void generateProxies(File outputDir,
			List<IBeanInfo> beans) throws ContextGeneratorException
	{
		try
		{
			for (IBeanInfo bean : beans)
			{
				if (bean.isProxyTarget())
				{
					generateProxy(outputDir, bean);
				}
			}
		}
		catch (IOException e)
		{
			throw new ContextGeneratorException(e);
		}
	}

	protected void generateProxy(File outputDir, IBeanInfo bean)
			throws IOException, ContextGeneratorException
	{
		File proxyFile = new File(outputDir,
				ProxyFactory.getProxyClassName(bean.getBeanClass()).replace(
						'.', File.separatorChar)
						+ AbstractJavaSourceContextGenerator.JAVA_EXTENSION);
		proxyFile.getParentFile().mkdirs();
		
		writer = new SourceWriter(new FileWriter(proxyFile));
		try
		{
			generateProxy(bean);
		}
		finally
		{
			writer.close();
			writer = null;
		}
	}

	protected void generateProxy(IBeanInfo bean) throws IOException,
			ContextGeneratorException
	{
		if (bean.getPackageName() != null)
		{
			writer.println("package " + bean.getPackageName() + ";");
			writer.println();
		}
		
		writer.println(getGeneratedOnComment());
		writer.println("public class "
				+ BeanUtils.getClassName(
						ProxyFactory.getProxyClassName(bean.getBeanClass()))
				+ " extends " + AbstractProxy.class.getName()
				+ " implements " + formatInterfaces(bean));
		writer.println("{");
		writer.indent();
		
		generateProxyClass(bean);
		Set<Method> methods = getInterfaceMethods(bean);
		generateArgArrays(bean, methods);
		generateMethods(bean, methods);
		
		writer.unindent();
		writer.println("}");
		writer.println();
	}

	protected void generateProxyClass(IBeanInfo bean)
	{
		writer.println("private static final "
				+ IProxyClass.class.getName() + " "
				+ PROXY_CLASS_NAME + " = new "
				+ ProxyFactory.getProxyMetaClassName(
						bean.getBeanClass()) + "();");
		writer.println();
	}

	protected void generateArgArrays(IBeanInfo bean, Set<Method> methods)
	{
		Set<Integer> lengths = new HashSet<Integer>();
		
		for (Method m : methods)
		{
			int length = m.getParameterTypes().length;
			
			if (!lengths.contains(length))
			{
				generateArgArray(length);
				lengths.add(length);
			}
		}
	}

	protected void generateArgArray(int length)
	{
		writer.println("private Object [] " + formatArgArrayName(length)
				+ " = new Object[" + length + "];");
		writer.println();
	}
	
	protected Set<Method> getInterfaceMethods(IBeanInfo bean)
			throws ContextGeneratorException
	{
		try
		{
			return BeanUtils.getInterfaceMethods(bean.getBeanClass());
		}
		catch (Exception e)
		{
			throw new ContextGeneratorException(e);
		}
	}

	protected void generateMethods(IBeanInfo bean,
			Collection<Method> methods)
	{
		int i = 0;
		for (Method m : methods)
		{
			generateMethod(bean, m, i++);
		}
	}

	protected void generateMethod(IBeanInfo bean, Method m, int index)
	{
		writer.println("private static final " + IMethodHandler.class.getName()
				+ " " + formatHandlerName(index)
				+ " = " + PROXY_CLASS_NAME + ".getMethodHandler("
				+ formatLiteral(m.getName()) + ", "
				+ formatClasses(m.getParameterTypes())
				+ ");");
		writer.println();
		
		writer.println("public synchronized "
				+ m.getReturnType().getName() + " "
				+ m.getName() + "(" + formatArguments(m) + ")");
		writer.println("{");
		writer.indent();
		
		writer.println("try");
		writer.println("{");
		writer.indent();
		
		generateArgArrayAssignments(m);
		writer.println((m.getReturnType() != Void.TYPE ? "return " : " ")
				+ formatReturnValue(m, index) + ";");
		
		writer.unindent();
		writer.println("}");
		generateCatchBlocks(m);
		
		writer.unindent();
		writer.println("}");
		writer.println();
	}

	protected void generateCatchBlocks(Method m)
	{
		for (Class cls : m.getExceptionTypes())
		{
			writer.println("catch (" + cls.getName() + " e)");
			writer.println("{");
			writer.println(SourceWriter.INDENT + "throw e;");
			writer.println("}");
		}
		
		if (!Arrays.asList(m.getExceptionTypes()).contains(Exception.class))
		{
			writer.println("catch (" + Exception.class.getName() + " e)");
			writer.println("{");
			/*writer.println(SourceWriter.INDENT
					+ "throw new " + RuntimeException.class.getName()
					+ "(e.toString());");*/
			writer.println(SourceWriter.INDENT + "throw wrap(e);");
			writer.println("}");
		}
	}

	protected void generateArgArrayAssignments(Method m)
	{
		final int length = m.getParameterTypes().length;
		
		for (int i = 0; i < length; i++)
		{
			writer.println(formatArgArrayName(length)
					+ "[" + i + "] = "
					+ formatWrapArgument(formatArgName(i),
							m.getParameterTypes()[i]) + ";");
		}
	}
}
