/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.aurorasoftworks.signal.tools.core.context.loader.spring;


import java.lang.reflect.Constructor;
import java.util.Hashtable;
import java.util.Stack;
import java.util.Vector;


import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.beans.BeanWrapper;
import org.springframework.beans.BeansException;
import org.springframework.beans.PropertyValues;
import org.springframework.beans.factory.BeanCreationException;
import org.springframework.beans.factory.BeanFactory;
import org.springframework.beans.factory.BeanInitializationException;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.beans.factory.support.DefaultListableBeanFactory;
import org.springframework.beans.factory.support.RootBeanDefinition;

import com.aurorasoftworks.signal.runtime.core.context.IBeanProcessor;
import com.aurorasoftworks.signal.runtime.core.context.IContext;
import com.aurorasoftworks.signal.runtime.core.context.IContextAware;
import com.aurorasoftworks.signal.runtime.core.context.IInitializingBean;
import com.aurorasoftworks.signal.tools.core.context.loader.BeanInfo;
import com.aurorasoftworks.signal.tools.core.context.loader.ContextReference;
import com.aurorasoftworks.signal.tools.core.context.loader.IBeanInfo;
import com.aurorasoftworks.signal.tools.core.context.loader.IBeanReference;
import com.aurorasoftworks.signal.tools.core.context.loader.IContextLoaderListener;


public class LoggingBeanFactory extends DefaultListableBeanFactory
		implements IBeanReferenceResolver
{
	private static final Log LOGGER
			= LogFactory.getLog(LoggingBeanFactory.class);
	
	private IContextLoaderListener listener;
	
	private Stack<IAutowiringState> autowiringState
			= new Stack<IAutowiringState>();

	private LoggingInstantiationStrategy instantiationStrategy;
	
	public LoggingBeanFactory(BeanFactory parentBeanFactory,
			IContextLoaderListener listener)
	{
		super(parentBeanFactory);
		
		this.listener = listener;
		instantiationStrategy = new LoggingInstantiationStrategy(
				listener, this);
		setInstantiationStrategy(instantiationStrategy);
		
		registerCustomEditor(Hashtable.class,
				HashtablePropertyEditor.getInstance());
		registerCustomEditor(Vector.class,
				VectorPropertyEditor.getInstance());
	}

	@Override
	public Object getBean(String name, Class requiredType)
			throws BeansException
	{
		return super.getBean(name, requiredType);
	}

	@Override
	public Object getBean(String name) throws BeansException
	{
		LOGGER.debug("getBean: " + name);
		Object result = super.getBean(name);
		saveBeanReference(name, result);

		return result;
	}

	@Override
	protected BeanWrapper autowireConstructor(String beanName,
			RootBeanDefinition mbd, Constructor [] ctors,
			Object[] explicitArgs)
	{
		LOGGER.debug("begin autowire: " + beanName);
		
		BeanWrapper result;
		startAutowiring();
		try
		{
			result = super.autowireConstructor(beanName, mbd,
					ctors,
					instantiationStrategy.resolveBeanReferences(
							explicitArgs));
		}
		finally
		{
			stopAutowiring();
			LOGGER.debug("end autowire");
		}
		
		return result;
	}

	protected void startAutowiring()
	{
		autowiringState.push(new AutowiringState());
	}
	
	protected void stopAutowiring()
	{
		autowiringState.pop();
	}

	@Override
	protected void applyPropertyValues(String beanName, BeanDefinition mbd,
			BeanWrapper bw, PropertyValues pvs)
	{
		LOGGER.debug("begin set properties: "
				+ pvs.getPropertyValues().length + ", " + beanName);
		
		startAutowiring();
		try
		{
			boolean isLazyInit = false;
			IBeanInfo info = instantiationStrategy.getBeanInfo(beanName);
			if (info != null)
			{
				isLazyInit = info.isLazyInit();
			}
			super.applyPropertyValues(beanName, mbd,
					new LoggingBeanWrapper(bw, beanName, isLazyInit,
							listener, this), pvs);
		}
		finally
		{
			stopAutowiring();
		}
		
		LOGGER.debug("end set properties");
	}
	
	protected void saveBeanReference(String name, Object object)
	{
		if (!autowiringState.isEmpty())
		{
			autowiringState.peek().saveBeanReference(
					name, object);
		}
	}

	public IBeanReference getBeanReference(Object object)
	{
		return autowiringState.peek()
				.getBeanReference(object);
	}
	
	public IBeanReference getBeanReference(String name)
	{
		return autowiringState.peek()
				.getBeanReference(name);
	}

	public boolean isBeanReference(Object object)
	{
		return ((!autowiringState.isEmpty()) &&
				(autowiringState.peek().isBeanReference(object)));
	}
	
	public boolean isBeanReference(String name)
	{
		return ((!autowiringState.isEmpty()) &&
				(autowiringState.peek().isBeanReference(name)));
	}
	
	@Override
	protected void invokeInitMethods(String beanName, Object bean,
			RootBeanDefinition mbd) throws Throwable
	{
		if ((bean instanceof IInitializingBean))
		{
			if (LOGGER.isDebugEnabled())
			{
				LOGGER.debug("Invoking afterPropertiesSet() on bean: " + beanName);
			}
			((IInitializingBean) bean).afterPropertiesSet();
			
			listener.onBeanMethodInvoked(createBeanInfo(beanName, bean),
					mbd.getBeanClass().getMethod("afterPropertiesSet"),
					LoggingInstantiationStrategy.EMPTY_ARGUMENT_ARRAY);
		}
	}

	protected IBeanInfo createBeanInfo(String beanName, Object bean)
	{
		IBeanInfo info = instantiationStrategy.getBeanInfo(beanName);
		if (info == null)
		{
			info = new BeanInfo(beanName, bean);
		}
		return info;
	}
	
	@Override
	protected Object initializeBean(String beanName, Object bean,
			RootBeanDefinition mbd)
	{
		Object result = super.initializeBean(beanName, bean, mbd);
		
		if ((bean instanceof IContextAware))
		{
			if (LOGGER.isDebugEnabled())
			{
				LOGGER.debug("Invoking setContext() on bean: " + beanName);
			}
			
			try
			{
				listener.onBeanMethodInvoked(createBeanInfo(beanName, bean),
						mbd.getBeanClass().getMethod("setContext", IContext.class),
						new Object[]{ContextReference.getInstance()});
			}
			catch (Exception e)
			{
				throw new BeanCreationException(
						"Failed to locate the setContext method", e);
			}
		}
		
		return result;
	}
	
	protected String[] getBeanPostProcessorNames()
	{
		return getBeanNamesForType(
				IBeanProcessor.class, true, false);
	}
	
	public Object applyBeanPostProcessorsBeforeInitialization(
			Object bean, String beanName) throws BeansException
	{
		String names [] = getBeanPostProcessorNames();
		Object result = bean;

		for (String processorName : names)
		{
			IBeanProcessor processor
					= (IBeanProcessor) getBean(processorName);
			try
			{
				/*result = processor.preProcessBean(
						beanName, result);*/
				listener.onBeanPreProcess(createBeanInfo(beanName, bean),
						createBeanInfo(processorName, processor));
			}
			catch (Exception e)
			{
				throw new BeanInitializationException(
						"postProcessBeforeInitialization failed", e);
			}
		}

		return result;
	}
	
	public Object applyBeanPostProcessorsAfterInitialization(
			Object bean, String beanName) throws BeansException
	{
		String names [] = getBeanPostProcessorNames();
		Object result = bean;
		
		for (String processorName : names)
		{
			IBeanProcessor processor
					= (IBeanProcessor) getBean(processorName);
			try
			{
				/*result = processor.postProcessBean(
						beanName, result);*/
				listener.onBeanPostProcess(createBeanInfo(beanName, bean),
						createBeanInfo(processorName, processor));
				
			}
			catch (Exception e)
			{
				throw new BeanInitializationException(
						"postProcessAfterInitialization failed", e);
			}
		}

		return result;
	}
}
