/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.aurorasoftworks.signal.tools.core.context.loader.spring;


import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;


import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.beans.factory.BeanCreationException;
import org.springframework.beans.factory.BeanFactory;
import org.springframework.beans.factory.support.RootBeanDefinition;
import org.springframework.beans.factory.support.SimpleInstantiationStrategy;

import com.aurorasoftworks.signal.tools.core.context.ContextProcessingException;
import com.aurorasoftworks.signal.tools.core.context.generator.ContextGeneratorException;
import com.aurorasoftworks.signal.tools.core.context.loader.BeanInfo;
import com.aurorasoftworks.signal.tools.core.context.loader.IBeanInfo;
import com.aurorasoftworks.signal.tools.core.context.loader.IContextLoaderListener;
import com.aurorasoftworks.signal.tools.core.util.BeanUtils;
import com.aurorasoftworks.signal.tools.core.util.MockUtils;


public class LoggingInstantiationStrategy extends SimpleInstantiationStrategy
{
	static final Object [] EMPTY_ARGUMENT_ARRAY = new Object[0];
	
	private static final Log LOGGER
			= LogFactory.getLog(LoggingInstantiationStrategy.class);
	
	private IContextLoaderListener listener;
	
	private IBeanReferenceResolver referenceResolver;
	
	private Map<String, IBeanInfo> beans;
	
	public LoggingInstantiationStrategy(IContextLoaderListener listener,
			IBeanReferenceResolver referenceResolver)
	{
		this.listener = listener;
		this.referenceResolver = referenceResolver;
		beans = new HashMap<String, IBeanInfo>();
	}
	
	@Override
	public Object instantiate(RootBeanDefinition beanDefinition,
			String beanName, BeanFactory owner, Constructor ctor,
			Object[] args)
	{
		LOGGER.debug("Creating bean ctor args: "
				+ beanDefinition.getBeanClassName());

		Object result = MockUtils.mock(ctor.getDeclaringClass());
		try
		{
			doInstantiate(beanDefinition, beanName, ctor, args, result);
		}
		catch (ContextGeneratorException e)
		{
			throw new BeanCreationException(
					"Failed to create bean: " + beanName, e);
		}
		
		return result;
	}

	private void doInstantiate(RootBeanDefinition beanDefinition,
			String beanName, Constructor ctor, Object[] args, Object bean)
			throws ContextGeneratorException
	{
		listener.onBeanInstantiated(createBeanInfo(beanName, bean,
				beanDefinition.isLazyInit()),
				ctor, resolveBeanReferences(args));
	}

	@Override
	public Object instantiate(RootBeanDefinition beanDefinition,
			String beanName, BeanFactory owner, Object factoryBean,
			Method factoryMethod, Object[] args)
	{
		LOGGER.debug("Creating bean factory args: " + beanDefinition.getBeanClassName());
		
		Object result = MockUtils.mock(factoryMethod.getReturnType());
		try
		{
			doInstantiate(beanDefinition, beanName, factoryBean, factoryMethod,
					args, result);
		}
		catch (ContextProcessingException e)
		{
			throw new BeanCreationException(
					"Failed to create bean: " + beanName, e);
		}
		
		return result;
	}

	private void doInstantiate(RootBeanDefinition beanDefinition,
			String beanName, Object factoryBean, Method factoryMethod,
			Object[] args, Object bean) throws ContextGeneratorException
	{
		if (factoryBean != null)
		{
			listener.onBeanInstantiated(
					createBeanInfo(beanName, bean,
							beanDefinition.isLazyInit()),
					new BeanInfo(beanDefinition.getFactoryBeanName(), factoryBean),
					factoryMethod, resolveBeanReferences(args));
		}
		else
		{
			listener.onBeanInstantiated(
					createBeanInfo(beanName, bean,
							beanDefinition.isLazyInit()),
					factoryMethod, args);
		}
	}

	@Override
	public Object instantiate(RootBeanDefinition beanDefinition,
			String beanName, BeanFactory owner)
	{
		LOGGER.debug("Creating bean: " + beanDefinition.getBeanClassName());
		Object result = MockUtils.mock(beanDefinition.getBeanClass());
		try
		{
			listener.onBeanInstantiated(
					createBeanInfo(beanName, result,
							beanDefinition.isLazyInit()), 
					result.getClass().getConstructor(new Class[]{}),
					EMPTY_ARGUMENT_ARRAY);
		}
		catch (Exception e)
		{
			throw new BeanCreationException(
					"Failed to find a default ctor for "
							+ result.getClass().getName(), e);
		}
	
		return result;
	}
	
	protected Object [] resolveBeanReferences(Object [] arguments)
	{
		Object [] result = null;
		
		if (arguments != null)
		{
			result = new Object[arguments.length];
			
			for (int i = 0; i < result.length; i++)
			{
				result[i] = resolveBeanReference(arguments[i]);
			}
		}

		return result;
	}
	
	protected Object resolveBeanReference(Object argument)
	{
		Object result = null;

		if (argument != null)
		{
			if (BeanUtils.isPrimitive(argument))
			{
				result = argument;
			}
			/* A collection might be either a bean reference or an inline
			 * <list>, <set> or <map>. Inline collections need to be treated
			 * in a different way */
			else if (BeanUtils.isCollection(argument))
			{
				if (referenceResolver.isBeanReference(argument))
				{
					result = referenceResolver.getBeanReference(argument);
				}
				else
				{
					if (argument instanceof List<?>)
					{
						result = resolveListBeanReferences(
								(List<Object>) argument);
					}
					else if (argument instanceof Set<?>)
					{
						result = resolveSetBeanReferences(
								(Set<Object>) argument);
					}
					else if (argument instanceof Map<?, ?>)
					{
						result = resolveMapBeanReferences(
								(Map<Object, Object>) argument);
					}
					else
					{
						result = argument;
					}
				}
			}
			else
			{
				result = referenceResolver.getBeanReference(argument);
			}
		}
		
		return result;
	}
	
	protected List<Object> resolveListBeanReferences(List<Object> beans)
	{
		List<?> copy = new ArrayList<Object>(beans);
		for (int i = 0; i < copy.size(); i++)
		{
			beans.set(i, resolveBeanReference(copy.get(i)));
		}
		
		return beans;
	}
	
	protected Set<Object> resolveSetBeanReferences(Set<Object> beans)
	{
		Set<?> copy = new HashSet<Object>(beans);
		for (Object element : copy)
		{
			if (referenceResolver.isBeanReference(element))
			{
				beans.remove(element);
				beans.add(resolveBeanReference(element));
			}
		}
		
		return beans;
	}
	
	protected Map<Object, Object> resolveMapBeanReferences(
			Map<Object, Object> beans)
	{
		Map<Object, Object> copy = new HashMap<Object, Object>(beans);
		for (Map.Entry<Object, Object> entry : copy.entrySet())
		{
			Object key = entry.getKey();
			Object value = entry.getValue();

			if (referenceResolver.isBeanReference(key))
			{
				beans.remove(key);
				key = resolveBeanReference(key);
			}
			
			if (referenceResolver.isBeanReference(value))
			{
				value = resolveBeanReference(value);
			}
			
			beans.put(key, value);
		}
		
		return beans;
	}
	
	protected IBeanInfo createBeanInfo(String beanName, Object bean,
			boolean lazyInit)
	{
		BeanInfo result = new BeanInfo(beanName, bean, lazyInit);
		beans.put(beanName, result);
		return result;
	}
	
	protected IBeanInfo createBeanInfo(String beanName, Object bean)
	{
		return createBeanInfo(beanName, bean, false);
	}
	
	public IBeanInfo getBeanInfo(String name)
	{
		return beans.get(name);
	}
}
