/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.aurorasoftworks.signal.tools.core.util;


import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;


public class BeanUtils
{
	private static final Set<Class<?>> WRAPPER_CLASSES
			= Collections.unmodifiableSet(new HashSet<Class<?>>()
			{{
				add(Integer.class);
				add(Long.class);
				add(Short.class);
				add(Float.class);
				add(Double.class);
				add(Boolean.class);
				add(Byte.class);
				add(Character.class);
				add(String.class);
			}});
	
	public static boolean isPrimitive(Object obj)
	{
		return ((obj != null)
				&& ((obj.getClass().isPrimitive()) 
						|| (WRAPPER_CLASSES.contains(obj.getClass()))));
	}
	
	public static boolean isCollection(Object obj)
	{
		return Collection.class.isAssignableFrom(obj.getClass())
				|| Map.class.isAssignableFrom(obj.getClass());
	}
	
	public static String getPackageName(String fullClassName)
	{
		String result;
		
		int index = fullClassName.lastIndexOf('.');
		if (index >= 0)
		{
			result = fullClassName.substring(0, index);
		}
		else
		{
			result = null;
		}
		
		return result;
	}
	
	public static String getClassName(String fullClassName)
	{
		String result;
		
		int index = fullClassName.lastIndexOf('.');
		if (index >= 0)
		{
			result = fullClassName.substring(index + 1);
		}
		else
		{
			result = fullClassName;
		}
		
		return result;
	}
	
	public static boolean isPublic(Method m)
	{
		return (m.getModifiers() & Modifier.PUBLIC) != 0;
	}
	
	public static Set<Method> getInterfaceMethods(Class<?> cls)
			throws SecurityException, NoSuchMethodException
	{
		return getInterfaceMethodMapping(cls).keySet();
	}

	public static Set<Class<?>> getAllClassInterfaces(Class<?> cls)
	{
		Set<Class<?>> intfaces = new HashSet<Class<?>>();
		Class<?> currentCls = cls;
		
		while(!Object.class.equals(currentCls))
		{
			intfaces.addAll(Arrays.asList(currentCls.getInterfaces()));
			currentCls = currentCls.getSuperclass();
		}

		return intfaces;
	}
	
	public static Map<Method, Class<?>> getInterfaceMethodMapping(
			Class<?> cls)
			throws SecurityException, NoSuchMethodException
	{
		Map<Method, Class<?>> mapping
				= new LinkedHashMap<Method, Class<?>>();

		for (Class<?> i : getAllClassInterfaces(cls))
		{
			for (Method method : i.getMethods())
			{
				Method methodImpl = cls.getMethod(method.getName(),
						method.getParameterTypes());

				if (!mapping.keySet().contains(methodImpl)
						&& (isPublic(methodImpl)))
				{
					mapping.put(methodImpl, i);
				}
			}
		}

		return mapping;
	}
	
	protected BeanUtils()
	{
	}
}
