/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.aurorasoftworks.signal.tools.core.context.loader.spring;


import static org.mockito.AdditionalMatchers.aryEq;
import static org.mockito.Matchers.argThat;
import static org.mockito.Matchers.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.verify;

import org.hamcrest.BaseMatcher;
import org.hamcrest.Description;
import org.hamcrest.Matcher;
import org.hamcrest.Matchers;
import org.springframework.context.ApplicationContext;
import org.testng.annotations.Test;

import com.aurorasoftworks.signal.runtime.core.context.IContext;
import com.aurorasoftworks.signal.tools.core.context.ContextProcessingException;
import com.aurorasoftworks.signal.tools.core.context.loader.Child;
import com.aurorasoftworks.signal.tools.core.context.loader.ContextAware;
import com.aurorasoftworks.signal.tools.core.context.loader.Eager;
import com.aurorasoftworks.signal.tools.core.context.loader.IBeanInfo;
import com.aurorasoftworks.signal.tools.core.context.loader.IBeanReference;
import com.aurorasoftworks.signal.tools.core.context.loader.IContextLoaderListener;
import com.aurorasoftworks.signal.tools.core.context.loader.Initializing;
import com.aurorasoftworks.signal.tools.core.context.loader.Lazy;
import com.aurorasoftworks.signal.tools.core.context.loader.Parent;
import com.aurorasoftworks.signal.tools.core.context.loader.PrimitiveArgs;
import com.aurorasoftworks.signal.tools.core.context.loader.PrimitiveProperties;
import com.aurorasoftworks.signal.tools.core.context.loader.PrimitivePropertiesProcessor;
import com.aurorasoftworks.signal.tools.core.context.loader.Simple;
import com.aurorasoftworks.signal.tools.core.context.loader.SimpleFactory;


@Test
public abstract class AbstractContextLoaderTest
{
	static class PropertyValue
	{
		private String name;
		
		private Object value;
		
		public PropertyValue(String name, Object value)
		{
			this.name = name;
			this.value = value;
		}

		public String getName()
		{
			return name;
		}

		public Object getValue()
		{
			return value;
		}
	}
	
	static abstract class AbstractMatcher<T> extends BaseMatcher<T>
	{
		public void describeTo(Description desc)
		{
		}
	}
	
	static class ChildReferenceMatcher extends AbstractMatcher<Object[]>
	{
		public boolean matches(Object arg)
		{
			Object [] a = (Object[]) arg;
			return (a.length == 1)
					&& (((IBeanReference) a[0]).getName().equals("child"));
		}
	}
	
	static class BeanReferenceMatcher extends AbstractMatcher<Object>
	{
		private String beanName;

		public BeanReferenceMatcher(String beanName)
		{
			this.beanName = beanName;
		}

		public boolean matches(Object arg)
		{
			return ((arg instanceof IBeanReference)
					&& ((IBeanReference) arg).getName().equals(beanName));
		}
	}
	
	static class BeanInfoMatcher extends AbstractMatcher<IBeanInfo>
	{
		private Matcher<String> nameMatcher;
		
		private Matcher<Object> beanMatcher;
		
		BeanInfoMatcher(String name, Class<?> beanClass)
		{
			nameMatcher = Matchers.equalTo(name);
			beanMatcher = Matchers.is(beanClass);
		}
		
		public boolean matches(Object arg)
		{
			boolean result = false;
			
			if (arg instanceof IBeanInfo)
			{
				IBeanInfo info = (IBeanInfo) arg;
				result = nameMatcher.matches(info.getName())
						&& beanMatcher.matches(info.getBean());
			}
			
			return result;
		}	
	}
	
	protected abstract org.springframework.context.ApplicationContext loadContext(String contextLocationBase,
			IContextLoaderListener listener)
			throws ContextProcessingException;

	public void testInstantiation() throws Exception
	{
		IContextLoaderListener listener = mock(IContextLoaderListener.class);
		
		loadContext(
				"/com/aurorasoftworks/signal/tools/core/context/loader/spring/simple.xml",
				listener);
		verify(listener).onBeanInstantiated(
				argThat(new BeanInfoMatcher("simple", Simple.class)),
				eq(Simple.class.getConstructor(new Class[]{})),
				aryEq(new Object[0]));
	}
	
	@Test(dependsOnMethods = "testInstantiation")
	public void testStaticFactoryInstantiation() throws Exception
	{
		IContextLoaderListener listener = mock(IContextLoaderListener.class);
		
		loadContext(
				"/com/aurorasoftworks/signal/tools/core/context/loader/spring/simple-static-factory.xml",
				listener);
		verify(listener).onBeanInstantiated(
				argThat(new BeanInfoMatcher("simple", Simple.class)),
				eq(Simple.class.getMethod("createInstance", (Class[]) null)),
				aryEq(new Object[0]));
	}
	
	@Test(dependsOnMethods = "testStaticFactoryInstantiation")
	public void testFactoryInstantiation() throws Exception
	{
		IContextLoaderListener listener = mock(IContextLoaderListener.class);
		
		loadContext(
				"/com/aurorasoftworks/signal/tools/core/context/loader/spring/simple-factory.xml",
				listener);
		verify(listener).onBeanInstantiated(
				argThat(new BeanInfoMatcher("simple-factory", SimpleFactory.class)),
				eq(SimpleFactory.class.getConstructor(new Class[]{})),
				aryEq(new Object[0]));
		verify(listener).onBeanInstantiated(
				argThat(new BeanInfoMatcher("simple", Simple.class)),
				argThat(new BeanInfoMatcher("simple-factory", SimpleFactory.class)),
				eq(SimpleFactory.class.getMethod("createInstance", (Class[]) null)),
				aryEq(new Object[0]));
	}
	
	@Test(dependsOnMethods = "testFactoryInstantiation")
	public void testInstantiationArgs() throws Exception
	{
		IContextLoaderListener listener = mock(IContextLoaderListener.class);
		
		loadContext(
				"/com/aurorasoftworks/signal/tools/core/context/loader/spring/primitive-args.xml",
				listener);
		verify(listener).onBeanInstantiated(
				argThat(new BeanInfoMatcher("args", PrimitiveArgs.class)),
				eq(PrimitiveArgs.class.getConstructors()[0]),
				aryEq(new Object[]{40000, 300000l, (short) 2000, (byte) 100, 'a', 1.0f, 20.0, "str"}));
	}
	
	@Test(dependsOnMethods = "testInstantiationArgs")
	public void testFactoryInstantiationArgs() throws Exception
	{
		IContextLoaderListener listener = mock(IContextLoaderListener.class);
		
		loadContext(
				"/com/aurorasoftworks/signal/tools/core/context/loader/spring/primitive-args-factory.xml",
				listener);
		verify(listener).onBeanInstantiated(
				argThat(new BeanInfoMatcher("args", PrimitiveArgs.class)),
				eq(PrimitiveArgs.class.getMethods()[0]),
				aryEq(new Object[]{40000, 300000l, (short) 2000, (byte) 100,
						'a', 1.0f, 20.0, "str"}));
	}
	
	@Test(dependsOnMethods = "testInstantiationArgs")
	public void testInstantiationProperties() throws Exception
	{
		IContextLoaderListener listener = mock(IContextLoaderListener.class);
		
		loadContext(
				"/com/aurorasoftworks/signal/tools/core/context/loader/spring/primitive-properties.xml",
				listener);
		verify(listener).onBeanInstantiated(
				argThat(new BeanInfoMatcher("props", PrimitiveProperties.class)),
				eq(PrimitiveProperties.class.getConstructors()[0]),
				aryEq(new Object[0]));
		
		PropertyValue [] pvs = new PropertyValue[]
		{
			new PropertyValue("i", 40000),
			new PropertyValue("l", 300000l),
			new PropertyValue("s", (short) 2000),
			new PropertyValue("b", (byte) 100),
			new PropertyValue("c", 'a'),
			new PropertyValue("f", 1.0f),
			new PropertyValue("d", 20.0),
			new PropertyValue("str", "str"),
		};
		
		for (PropertyValue pv : pvs)
		{
			verify(listener).onBeanPropertySet(
					argThat(new BeanInfoMatcher("props", PrimitiveProperties.class)),
					eq(pv.getName()),
					eq(pv.getValue()));
		}
	}
	
	@Test(dependsOnMethods = {"testInstantiationArgs",
			"testInstantiationProperties"})
	public void testInstantiationWithReferenceArg() throws Exception
	{
		IContextLoaderListener listener = mock(IContextLoaderListener.class);
		
		loadContext(
				"/com/aurorasoftworks/signal/tools/core/context/loader/spring/reference-arg.xml",
				listener);
		
		verify(listener).onBeanInstantiated(
				argThat(new BeanInfoMatcher("child", Child.class)),
				eq(Child.class.getConstructors()[0]),
				aryEq(new Object[0]));
		verify(listener).onBeanInstantiated(
				argThat(new BeanInfoMatcher("parent", Parent.class)),
				eq(Parent.class.getConstructors()[0]),
				argThat(new ChildReferenceMatcher()));
	}
	
	@Test(dependsOnMethods = "testInstantiationWithReferenceArg")
	public void testInstantiationWithReferenceProperty() throws Exception
	{
		IContextLoaderListener listener = mock(IContextLoaderListener.class);

		loadContext(
				"/com/aurorasoftworks/signal/tools/core/context/loader/spring/reference-property.xml",
				listener);
		verify(listener).onBeanInstantiated(
				argThat(new BeanInfoMatcher("child", Child.class)),
				eq(Child.class.getConstructors()[0]),
				aryEq(new Object[0]));
		verify(listener).onBeanInstantiated(
				argThat(new BeanInfoMatcher("parent", Parent.class)),
				eq(Parent.class.getConstructors()[0]),
				aryEq(new Object[]{null}));
		verify(listener).onBeanPropertySet(
				argThat(new BeanInfoMatcher("parent", Parent.class)),
				eq("child2"),
				argThat(new BeanReferenceMatcher("child")));
	}
	
	@Test(dependsOnMethods = "testInstantiation")
	public void testInitialization() throws Exception
	{
		IContextLoaderListener listener = mock(IContextLoaderListener.class);
		
		loadContext(
				"/com/aurorasoftworks/signal/tools/core/context/loader/spring/initializing.xml",
				listener);
		
		listener.onBeanMethodInvoked(
				argThat(new BeanInfoMatcher("init", Initializing.class)),
				eq(Initializing.class.getMethod("afterPropertiesSet")),
				aryEq(new Object[]{}));
	}
	
	@Test(dependsOnMethods = "testInitialization")
	public void testContextAware() throws Exception
	{
		IContextLoaderListener listener = mock(IContextLoaderListener.class);
		
		loadContext(
				"/com/aurorasoftworks/signal/tools/core/context/loader/spring/context-aware.xml",
				listener);
		
		listener.onBeanMethodInvoked(
				argThat(new BeanInfoMatcher("contextAware", ContextAware.class)),
				eq(ContextAware.class.getMethod("setContext", IContext.class)),
				aryEq(new Object[]{}));
	}
	
	@Test(dependsOnMethods = "testInitialization")
	public void testLazyInit() throws Exception
	{
		IContextLoaderListener listener = mock(IContextLoaderListener.class);
		
		Eager.resetInstanceCount();
		Lazy.resetInstanceCount();
		
		ApplicationContext ctx = loadContext(
				"/com/aurorasoftworks/signal/tools/core/context/loader/spring/lazy-init.xml",
				listener);
		ctx.getBean("lazy");
	}
	
	@Test(dependsOnMethods = "testFactoryInstantiation")
	public void testBeanProcessor() throws Exception
	{
		IContextLoaderListener listener = mock(IContextLoaderListener.class);
		
		loadContext(
				"/com/aurorasoftworks/signal/tools/core/context/loader/spring/post-processor.xml",
				listener);

		verify(listener).onBeanPreProcess(
				argThat(new BeanInfoMatcher("props", PrimitiveProperties.class)),
				argThat(new BeanInfoMatcher("post-processor", PrimitivePropertiesProcessor.class)));
		verify(listener).onBeanPostProcess(
				argThat(new BeanInfoMatcher("props", PrimitiveProperties.class)),
				argThat(new BeanInfoMatcher("post-processor", PrimitivePropertiesProcessor.class)));
	}
}
