/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.aurorasoftworks.signal.tools.core.maven;


import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;


import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.project.MavenProject;
import org.springframework.context.ApplicationContext;

import com.aurorasoftworks.signal.tools.core.context.ContextProcessingException;
import com.aurorasoftworks.signal.tools.core.context.generator.source.java.JavaSourceContextGenerator;
import com.aurorasoftworks.signal.tools.core.context.loader.spring.FileSystemContextLoader;


/**
 * A Maven plugin that can generate contexts from Spring
 * configurations files. This plugin is implemented as a thin wrapper
 * for {@link JavaSourceContextGenerator}.
 * 
 * @author Marek
 * 
 * @goal generate-spring-context
 */
public class SpringContextPlugin extends AbstractMojo
{	
	/**
     * Location of the file.
     * 
     * @parameter expression="${project.build.directory}/generated-sources/signal"
     * @required
     */
    private File outputDirectory;
	
    /**
     * 
     * @parameter
     * @required
     */
	private File contextFile;

	/**
	 * @parameter
	 * @required
	 */
	private String targetClassName;
	
	/**
	 * @parameter expression="com.aurorasoftworks.signal.runtime.core.context.Context"
	 */
	private String baseClassName;
	
	/**
	 * @parameter expression="true"
	 */
	private boolean generateProxyClasses;
	
	/**
	 * @parameter expression="${basedir}"
	 */
	private File basedir;

	/**
	 * @parameter expression="${project}"
	 * @required
     * @readonly
	 */
	private MavenProject project;

	public void execute() throws MojoExecutionException, MojoFailureException
	{		
		getLog().info("Generating a Spring context");
		
		getLog().info("Working directory: " + System.getProperty("user.dir"));
		getLog().info("Output directory: " + outputDirectory);
		getLog().info("Context input file: " + contextFile);
		getLog().info("Context output class: " + targetClassName);
		getLog().debug("Base directory: " + basedir);
		getLog().debug("Generate proxy classes: " + generateProxyClasses);

		long t0 = System.currentTimeMillis();
		
		try
		{
			generateContext();
		}
		catch (Exception e)
		{
			throw new MojoExecutionException(
					"Failed to process the context", e);
		}
		
		getLog().info("Spring context generated in "
				+ (System.currentTimeMillis() - t0) + " ms");
	}

	protected void generateContext() throws ContextProcessingException,
			IOException, ClassNotFoundException
	{
		/* 
		 * target/classes needs to be added to the classpath so that compiled
		 * classes are visible.
		 */
		URL url = new File(basedir, "target" + File.separator + "classes")
				.toURI().toURL();
		getLog().debug("Classloader URL: " + url);
		
		ClassLoader classLoader = new URLClassLoader(new URL[]{url},
					getClass().getClassLoader());
		Thread.currentThread().setContextClassLoader(classLoader);
		
		/* The File is converted to URI to correctly handle spaces */
		String contextLocation = contextFile.toURI().toString();

		getLog().info("Context location: " + contextLocation);

		JavaSourceContextGenerator<ApplicationContext> generator
				= new JavaSourceContextGenerator<ApplicationContext>();
		generator.setContextBaseClassFullName(baseClassName);
		generator.setClassLoader(classLoader);
		generator.setGenerateProxyClasses(generateProxyClasses);

		generator.generateContext(contextLocation,
				targetClassName,
				FileSystemContextLoader.getInstance(),
				outputDirectory);
		project.addCompileSourceRoot(outputDirectory.getPath());
	}
}
